@@ -4,6 +4,7 @@ from feeluown.excs import ModelNotFound
 from feeluown.library import (
     AbstractProvider,
     BriefSongModel,
+    BriefPlaylistModel,
     PlaylistModel,
     Collection,
     CollectionType,
@@ -23,12 +24,17 @@ from feeluown.library import (
     SupportsPlaylistGet,
     SupportsPlaylistSongsReader,
     SupportsRecACollectionOfSongs,
+    SupportsCurrentUserDislikeSongsReader,
+    SupportsCurrentUserDislikeAddSong,
+    SupportsCurrentUserDislikeRemoveSong,
+    SupportsCurrentUserChanged,
     SimpleSearchResult,
     SearchType,
     ModelType,
     UserModel,
 )
 from feeluown.media import Media, Quality
+from feeluown.utils.dispatch import Signal
 from feeluown.utils.reader import create_reader, SequentialReader
 from .api import API
 from .login import read_cookies
@@ -53,6 +59,9 @@ class Supports(
     SupportsPlaylistSongsReader,
     SupportsRecACollectionOfSongs,
     SupportsAlbumSongsReader,
+    SupportsCurrentUserDislikeSongsReader,
+    SupportsCurrentUserDislikeAddSong,
+    SupportsCurrentUserDislikeRemoveSong,
     Protocol,
 ):
     pass
@@ -70,6 +79,7 @@ class QQProvider(AbstractProvider, ProviderV2):
     def __init__(self):
         super().__init__()
         self.api = API()
+        self.current_user_changed = Signal()
 
     def _(self) -> Supports:
         return self
@@ -89,6 +99,7 @@ class QQProvider(AbstractProvider, ProviderV2):
             self.auth(user)
         else:
             logger.info(f'Auto login failed: {err}')
+        self.current_user_changed.emit(user)
 
     def try_get_user_from_cookies(self, cookies) -> Tuple[Optional[UserModel], str]:
         if not cookies:  # is None or empty
@@ -172,7 +183,7 @@ class QQProvider(AbstractProvider, ProviderV2):
         return Media(q_media_mapping[quality.value])
 
     def video_list_quality(self, video):
-        q_media_mapping = self._model_cache_get_or_fetch(video, "q_url_mapping")
+        q_media_mapping = self._model_cache_get_or_fetch(video, "q_url_mapping", ttl=3600)
         return [Quality.Video(k) for k, v in q_media_mapping.items() if v]
 
     def song_list_quality(self, song) -> List[Quality.Audio]:
@@ -212,9 +223,11 @@ class QQProvider(AbstractProvider, ProviderV2):
         q_media_mapping, exists = song.cache_get("q_media_mapping")
         if exists is True:
             return q_media_mapping
-        quality_suffix = self._model_cache_get_or_fetch(song, "quality_suffix")
-        mid = self._model_cache_get_or_fetch(song, "mid")
-        media_id = self._model_cache_get_or_fetch(song, "media_id")
+        # 根据过去经验，这个东西一段时间之后会过期。
+        # 具体多久，不是很清楚。
+        quality_suffix = self._model_cache_get_or_fetch(song, "quality_suffix", ttl=3600)
+        mid = self._model_cache_get_or_fetch(song, "mid", ttl=3600)
+        media_id = self._model_cache_get_or_fetch(song, "media_id", ttl=3600)
         q_media_mapping = {}
         # 注：self.quality_suffix 这里可能会触发一次网络请求
         for idx, (q, t, b, s) in enumerate(quality_suffix):
@@ -270,6 +283,22 @@ class QQProvider(AbstractProvider, ProviderV2):
         data = self.api.playlist_detail(int(identifier), limit=1000)
         return _deserialize(data, QQPlaylistSchema)
 
+    def playlist_add_song(self, playlist, song):
+        # FIXME: 目前 playlist 相关接口用的都是 diss 结构体，而这里需要一个 dirid。
+        # 平台方也提供了 dir 相关的接口，我大胆猜测，diss 是一套老接口。
+        playlist._cache.pop("songs", None)
+        dirid = self._get_dirid_by_playlist_id(playlist.identifier)
+        return self.api.playlist_add_songs(dirid, [song.identifier])
+
+    def playlist_remove_song(self, playlist, song):
+        playlist._cache.pop("songs", None)
+        dirid = self._get_dirid_by_playlist_id(playlist.identifier)
+        return self.api.playlist_remove_songs(dirid, [song.identifier])
+
+    def _get_dirid_by_playlist_id(self, playlist_id):
+        data = self.api.playlist_detail(int(playlist_id), limit=1)
+        return data["dirid"]
+
     def playlist_create_songs_rd(self, playlist):
         songs = self._model_cache_get_or_fetch(playlist, "songs")
         return create_reader(songs)
@@ -371,7 +400,13 @@ class QQProvider(AbstractProvider, ProviderV2):
         if user is None:
             return []
         playlists = self._model_cache_get_or_fetch(user, "playlists")
-        return playlists
+        fav_pid = self._model_cache_get_or_fetch(user, "fav_pid")
+        my_love = BriefPlaylistModel(
+            source=SOURCE,
+            identifier=str(fav_pid),
+            name='我喜欢'
+        )
+        return [my_love] + playlists
 
     def current_user_fav_create_songs_rd(self):
         user = self.get_current_user()
@@ -405,6 +440,9 @@ class QQProvider(AbstractProvider, ProviderV2):
             return create_reader([])
         mid = self._model_cache_get_or_fetch(user, "mid")
         playlists = self.api.user_favorite_playlists(user.identifier, mid)
+        # HACK: 给 playlist 加一个 disstid 字段，这样可以兼容 QQPlaylistSchema
+        for playlist in playlists:
+            playlist["disstid"] = playlist["dissid"]
         return [_deserialize(playlist, QQPlaylistSchema) for playlist in playlists]
 
     def has_current_user(self):
@@ -417,6 +455,35 @@ class QQProvider(AbstractProvider, ProviderV2):
         data_songs = self.api.song_similar(int(song.identifier))
         return [_deserialize(data_song, QQSongSchema) for data_song in data_songs]
 
+    def current_user_dislike_create_songs_rd(self):
+        user = self.get_current_user()
+        if user is None:
+            return create_reader([])
+        # FIXME: 如果用户的黑名单歌曲数量较多的话，这样处理则是不够的
+        items = self.api.get_dislike_list(1, API.DislikeListType.song, 0)
+        songs = []
+        for item in items:
+            name = item['Name']
+            title, artists_name = name.split(' - ')
+            song = BriefSongModel(
+                source=SOURCE,
+                identifier=item['ID'],
+                title=title,
+                artists_name=artists_name,
+            )
+            songs.append(song)
+        return create_reader(songs)
+
+    def current_user_dislike_add_song(self, song):
+        items = [{'ID': song.identifier}]
+        js = self.api.add_to_dislike_list(items, API.DislikeListType.song)
+        return js.get('Retcode') == 0
+
+    def current_user_dislike_remove_song(self, song):
+        items = [{'ID': song.identifier}]
+        js = self.api.remove_from_dislike_list(items, API.DislikeListType.song)
+        return js.get('Retcode') == 0
+
 
 def _deserialize(data, schema_cls):
     schema = schema_cls()
